import os
from typing import List

import click
from jinja2 import Template
from stellar_sdk import __version__ as stellar_sdk_version, StrKey
from stellar_sdk import xdr

from stellar_contract_bindings import __version__ as stellar_contract_bindings_version
from stellar_contract_bindings.utils import get_specs_by_contract_id


def is_keywords(word: str) -> bool:
    return word in [
        "abstract",
        "as",
        "assert",
        "async",
        "await",
        "break",
        "case",
        "catch",
        "class",
        "const",
        "continue",
        "covariant",
        "default",
        "deferred",
        "do",
        "dynamic",
        "else",
        "enum",
        "export",
        "extends",
        "extension",
        "external",
        "factory",
        "false",
        "final",
        "finally",
        "for",
        "Function",
        "get",
        "hide",
        "if",
        "implements",
        "import",
        "in",
        "interface",
        "is",
        "late",
        "library",
        "mixin",
        "new",
        "null",
        "on",
        "operator",
        "part",
        "rethrow",
        "required",
        "return",
        "set",
        "show",
        "static",
        "super",
        "switch",
        "sync",
        "this",
        "throw",
        "true",
        "try",
        "typedef",
        "var",
        "void",
        "while",
        "with",
        "yield",
    ]


def is_tuple_struct(entry: xdr.SCSpecUDTStructV0) -> bool:
    return all(f.name.isdigit() for f in entry.fields)


def prefixed_type_name(type_name: str, class_name: str) -> str:
    """Prefix a type name with the class name to avoid conflicts.
    
    Args:
        type_name: The original type name from the contract spec
        class_name: The class name to use as prefix
    
    Returns:
        The prefixed type name if it's a UDT, or the original if it's a built-in type
    """
    # Don't prefix built-in types
    if type_name in ['bool', 'int', 'BigInt', 'String', 'Uint8List', 'Address', 'Map', 'List', 
                     'XdrSCVal', 'void']:
        return type_name
    return f"{class_name}{type_name}"


def snake_to_camel(text: str, first_letter_lower: bool = True) -> str:
    parts = text.split("_")
    if first_letter_lower:
        return parts[0].lower() + "".join(part.capitalize() for part in parts[1:])
    else:
        return "".join(part.capitalize() for part in parts)

def camel_to_snake(text: str) -> str:
    result = text[0].lower()
    for char in text[1:]:
        if char.isupper():
            result += "_" + char.lower()
        else:
            result += char
    return result

def to_dart_type(td: xdr.SCSpecTypeDef, nullable: bool = False, class_name: str = "") -> str:
    t = td.type
    nullable_suffix = "?" if nullable else ""

    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"XdrSCVal{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"bool{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return "void"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"int{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"BigInt{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"BigInt{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"BigInt{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"BigInt{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"Uint8List{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"String{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"String{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"Address{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"Address{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return to_dart_type(td.option.value_type, nullable=True, class_name=class_name)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return to_dart_type(ok_t, nullable, class_name=class_name)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return f"List<{to_dart_type(td.vec.element_type, class_name=class_name)}>{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"Map<{to_dart_type(td.map.key_type, class_name=class_name)}, {to_dart_type(td.map.value_type, class_name=class_name)}>{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "void"
        types = [to_dart_type(t, class_name=class_name) for t in td.tuple.value_types]
        return f"({', '.join(types)}){nullable_suffix}"  # Using Dart 3 records
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"Uint8List{nullable_suffix}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        udt_name = td.udt.name.decode()
        return f"{prefixed_type_name(udt_name, class_name)}{nullable_suffix}"
    raise ValueError(f"Unsupported SCValType: {t}")


def to_scval(td: xdr.SCSpecTypeDef, name: str, class_name: str = "") -> str:
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"{name}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"XdrSCVal.forBool({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"XdrSCVal.forVoid()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"XdrSCVal.forU32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"XdrSCVal.forI32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"XdrSCVal.forU64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"XdrSCVal.forI64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"XdrSCVal.forTimePoint({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"XdrSCVal.forDuration({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"XdrSCVal.forU128BigInt({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"XdrSCVal.forI128BigInt({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"XdrSCVal.forU256BigInt({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"XdrSCVal.forI256BigInt({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"XdrSCVal.forBytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"XdrSCVal.forString({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"XdrSCVal.forSymbol({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"{name}.toXdrSCVal()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"{name}.toXdrSCVal()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return f"{name} == null ? XdrSCVal.forVoid() : {to_scval(td.option.value_type, name, class_name)}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        return NotImplementedError("SC_SPEC_TYPE_RESULT is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return f"XdrSCVal.forVec({name}.map((e) => {to_scval(td.vec.element_type, 'e', class_name)}).toList())"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"XdrSCVal.forMap(Map.fromEntries({name}.entries.map((e) => MapEntry({to_scval(td.map.key_type, 'e.key', class_name)}, {to_scval(td.map.value_type, 'e.value', class_name)}))))"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        types = [
            to_scval(t, f"{name}.${{i+1}}", class_name) for i, t in enumerate(td.tuple.value_types)
        ]
        return f"XdrSCVal.forVec([{', '.join(types)}])"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"XdrSCVal.forBytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        return f"{name}.toScVal()"
    raise ValueError(f"Unsupported SCValType: {t}")


def from_scval(td: xdr.SCSpecTypeDef, name: str, class_name: str = "") -> str:
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"{name}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"{name}.b!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"null"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"{name}.u32!.uint32"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"{name}.i32!.int32"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"BigInt.from({name}.u64!.uint64)"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"{name}.i64!.int64"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"BigInt.from({name}.u64!.uint64)"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"BigInt.from({name}.u64!.uint64)"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"{name}.toBigInt()!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"{name}.toBigInt()!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"{name}.toBigInt()!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"{name}.toBigInt()!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"{name}.bytes!.dataValue"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"{name}.str!"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"{name}.sym!.toString()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"Address.fromXdrSCVal({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"Address.fromXdrSCVal({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return f"{name}.discriminant == XdrSCValType.SCV_VOID ? null : {from_scval(td.option.value_type, name, class_name)}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return f"{from_scval(ok_t, name, class_name)}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return (
            f"{name}.vec!.map((e) => {from_scval(td.vec.element_type, 'e', class_name)}).toList()"
        )
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"Map.fromEntries({name}.map!.entries.map((e) => MapEntry({from_scval(td.map.key_type, 'e.key', class_name)}, {from_scval(td.map.value_type, 'e.val', class_name)})))"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "null"
        elements = f"{name}.vec!"
        types = [
            from_scval(t, f"{elements}[{i}]", class_name)
            for i, t in enumerate(td.tuple.value_types)
        ]
        return f"({', '.join(types)})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"{name}.bytes!.dataValue"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        udt_name = td.udt.name.decode()
        return f"{prefixed_type_name(udt_name, class_name)}.fromScVal({name})"
    raise NotImplementedError(f"Unsupported SCValType: {t}")


def render_info():
    return f"// This file was generated by stellar_contract_bindings v{stellar_contract_bindings_version} and stellar_sdk v{stellar_sdk_version}."


def render_imports():
    template = """
import 'dart:typed_data';
import 'package:stellar_flutter_sdk/stellar_flutter_sdk.dart';
"""
    return Template(template).render()


def render_enum(entry: xdr.SCSpecUDTEnumV0, class_name: str):
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/// {{ entry.doc.decode() if entry.doc else type_name + ' enum' }}
enum {{ type_name }} {
  {%- for case in entry.cases %}
  {{ snake_to_camel(case.name.decode(), False) }}({{ case.value.uint32 }}){% if loop.last %};{% else %},{% endif %}
  {%- endfor %}

  final int value;
  
  const {{ type_name }}(this.value);
  
  factory {{ type_name }}.fromValue(int value) {
    return {{ type_name }}.values.firstWhere(
      (e) => e.value == value,
      orElse: () => throw ArgumentError('Unknown {{ type_name }} value: $value'),
    );
  }
  
  XdrSCVal toScVal() {
    return XdrSCVal.forU32(value);
  }
  
  static {{ type_name }} fromScVal(XdrSCVal val) {
    return {{ type_name }}.fromValue(val.u32!.uint32);
  }
}
"""
    rendered_code = Template(template).render(
        entry=entry, type_name=type_name, snake_to_camel=snake_to_camel
    )
    return rendered_code


def render_error_enum(entry: xdr.SCSpecUDTErrorEnumV0, class_name: str):
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/// {{ entry.doc.decode() if entry.doc else type_name + ' error enum' }}
enum {{ type_name }} {
  {%- for case in entry.cases %}
  {{ snake_to_camel(case.name.decode(), False) }}({{ case.value.uint32 }}){% if loop.last %};{% else %},{% endif %}
  {%- endfor %}

  final int value;
  
  const {{ type_name }}(this.value);
  
  factory {{ type_name }}.fromValue(int value) {
    return {{ type_name }}.values.firstWhere(
      (e) => e.value == value,
      orElse: () => throw ArgumentError('Unknown {{ type_name }} value: $value'),
    );
  }
  
  XdrSCVal toScVal() {
    return XdrSCVal.forU32(value);
  }
  
  static {{ type_name }} fromScVal(XdrSCVal val) {
    return {{ type_name }}.fromValue(val.u32!.uint32);
  }
}
"""
    rendered_code = Template(template).render(
        entry=entry, type_name=type_name, snake_to_camel=snake_to_camel
    )
    return rendered_code


def render_struct(entry: xdr.SCSpecUDTStructV0, class_name: str):
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    
    # Create wrapper functions with class_name bound
    def to_dart_type_bound(td, nullable=False):
        return to_dart_type(td, nullable, class_name)
    
    def to_scval_bound(td, name):
        return to_scval(td, name, class_name)
    
    def from_scval_bound(td, name):
        return from_scval(td, name, class_name)
    
    template = """
/// {{ entry.doc.decode() if entry.doc else type_name + ' struct' }}
class {{ type_name }} {
  {%- for field in entry.fields %}
  final {{ to_dart_type(field.type) }} {{ snake_to_camel(field.name.decode()) }};
  {%- endfor %}

  const {{ type_name }}({
    {%- for field in entry.fields %}
    required this.{{ snake_to_camel(field.name.decode()) }},
    {%- endfor %}
  });

  XdrSCVal toScVal() {
    final fields = <XdrSCMapEntry>[];
    {%- for field in entry.fields %}
    fields.add(XdrSCMapEntry(
      XdrSCVal.forSymbol('{{ field.name_r.decode() if field.name_r else field.name.decode() }}'),
      {{ to_scval(field.type, snake_to_camel(field.name.decode())) }},
    ));
    {%- endfor %}
    return XdrSCVal.forMap(fields);
  }

  factory {{ type_name }}.fromScVal(XdrSCVal val) {
    final map = val.map!;
    final fieldsMap = <String, XdrSCVal>{};
    for (final entry in map) {
      fieldsMap[entry.key.sym!.toString()] = entry.val;
    }
    
    return {{ type_name }}(
      {%- for field in entry.fields %}
      {{ snake_to_camel(field.name.decode()) }}: {{ from_scval(field.type, 'fieldsMap["' ~ (field.name_r.decode() if field.name_r else field.name.decode()) ~ '"]!') }},
      {%- endfor %}
    );
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is {{ type_name }} &&
          {%- for field in entry.fields %}
          {{ snake_to_camel(field.name.decode()) }} == other.{{ snake_to_camel(field.name.decode()) }}{% if not loop.last %} &&{% endif %}
          {%- endfor %};

  @override
  int get hashCode => Object.hash(
      {%- for field in entry.fields %}
      {{ snake_to_camel(field.name.decode()) }}{% if not loop.last %},{% endif %}
      {%- endfor %}
  );
}
"""
    rendered_code = Template(template).render(
        entry=entry,
        type_name=type_name,
        to_dart_type=to_dart_type_bound,
        to_scval=to_scval_bound,
        from_scval=from_scval_bound,
        snake_to_camel=snake_to_camel,
    )
    return rendered_code


def render_tuple_struct(entry: xdr.SCSpecUDTStructV0, class_name: str):
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    
    # Create wrapper functions with class_name bound
    def to_dart_type_bound(td, nullable=False):
        return to_dart_type(td, nullable, class_name)
    
    def to_scval_bound(td, name):
        return to_scval(td, name, class_name)
    
    def from_scval_bound(td, name):
        return from_scval(td, name, class_name)
    
    template = """
/// {{ entry.doc.decode() if entry.doc else type_name + ' tuple struct' }}
class {{ type_name }} {
  final ({% for f in entry.fields %}{{ to_dart_type(f.type) }}{% if not loop.last %}, {% endif %}{% endfor %}) value;

  const {{ type_name }}(this.value);

  XdrSCVal toScVal() {
    return XdrSCVal.forVec([
      {%- for f in entry.fields %}
      {{ to_scval(f.type, 'value.$' + (loop.index0 + 1)|string) }}{% if not loop.last %},{% endif %}
      {%- endfor %}
    ]);
  }

  factory {{ type_name }}.fromScVal(XdrSCVal val) {
    final vec = val.vec!;
    return {{ type_name }}((
      {%- for f in entry.fields %}
      {{ from_scval(f.type, 'vec[' + loop.index0|string + ']') }}{% if not loop.last %},{% endif %}
      {%- endfor %}
    ));
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is {{ type_name }} && value == other.value;

  @override
  int get hashCode => value.hashCode;
}
"""
    rendered_code = Template(template).render(
        entry=entry,
        type_name=type_name,
        to_dart_type=to_dart_type_bound,
        to_scval=to_scval_bound,
        from_scval=from_scval_bound
    )
    return rendered_code


def render_union(entry: xdr.SCSpecUDTUnionV0, class_name: str):
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    
    # Create wrapper functions with class_name bound
    def to_dart_type_bound(td, nullable=False):
        return to_dart_type(td, nullable, class_name)
    
    def to_scval_bound(td, name):
        return to_scval(td, name, class_name)
    
    def from_scval_bound(td, name):
        return from_scval(td, name, class_name)
    
    kind_enum_template = """
/// Kind enum for {{ type_name }}
enum {{ type_name }}Kind {
  {%- for case in entry.cases %}
  {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
  {{ snake_to_camel(case.void_case.name.decode(), False) }}('{{ case.void_case.name_r.decode() if case.void_case.name_r else case.void_case.name.decode() }}'){% if loop.last %};{% else %},{% endif %}
  {%- else %}
  {{ snake_to_camel(case.tuple_case.name.decode(), False) }}('{{ case.tuple_case.name.decode() if case.tuple_case.name_r else case.tuple_case.name.decode() }}'){% if loop.last %};{% else %},{% endif %}
  {%- endif %}
  {%- endfor %}

  final String value;
  
  const {{ type_name }}Kind(this.value);
  
  factory {{ type_name }}Kind.fromValue(String value) {
    return {{ type_name }}Kind.values.firstWhere(
      (e) => e.value == value,
      orElse: () => throw ArgumentError('Unknown {{ type_name }}Kind value: $value'),
    );
  }
}
"""
    kind_enum_rendered_code = Template(kind_enum_template).render(
        entry=entry, type_name=type_name, xdr=xdr, snake_to_camel=snake_to_camel
    )

    template = """
/// {{ entry.doc.decode() if entry.doc else type_name + ' union' }}
class {{ type_name }} {
  final {{ type_name }}Kind kind;
  {%- for case in entry.cases %}
  {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
  {%- if len(case.tuple_case.type) == 1 %}
  final {{ to_dart_type(case.tuple_case.type[0], nullable=True) }} {{ snake_to_camel(case.tuple_case.name.decode()) }};
  {%- else %}
  final ({% for f in case.tuple_case.type %}{{ to_dart_type(f) }}{% if not loop.last %}, {% endif %}{% endfor %})? {{ snake_to_camel(case.tuple_case.name.decode()) }};
  {%- endif %}
  {%- endif %}
  {%- endfor %}

  const {{ type_name }}._({
    required this.kind,
    {%- for case in entry.cases %}
    {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
    this.{{ snake_to_camel(case.tuple_case.name.decode()) }},
    {%- endif %}
    {%- endfor %}
  });

  {%- for case in entry.cases %}
  {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
  factory {{ type_name }}.{{ snake_to_camel(case.void_case.name.decode()) }}() {
    return {{ type_name }}._(kind: {{ type_name }}Kind.{{ snake_to_camel(case.void_case.name.decode(), False) }});
  }
  {%- else %}
  {%- if len(case.tuple_case.type) == 1 %}
  factory {{ type_name }}.{{ snake_to_camel(case.tuple_case.name.decode()) }}({{ to_dart_type(case.tuple_case.type[0]) }} value) {
    return {{ type_name }}._(
      kind: {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }},
      {{ snake_to_camel(case.tuple_case.name.decode()) }}: value,
    );
  }
  {%- else %}
  factory {{ type_name }}.{{ snake_to_camel(case.tuple_case.name.decode()) }}(({% for f in case.tuple_case.type %}{{ to_dart_type(f) }}{% if not loop.last %}, {% endif %}{% endfor %}) value) {
    return {{ type_name }}._(
      kind: {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }},
      {{ snake_to_camel(case.tuple_case.name.decode()) }}: value,
    );
  }
  {%- endif %}
  {%- endif %}
  {%- endfor %}

  XdrSCVal toScVal() {
    switch (kind) {
      {%- for case in entry.cases %}
      {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.void_case.name.decode(), False) }}:
        return XdrSCVal.forVec([XdrSCVal.forSymbol(kind.value)]);
      {%- else %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }}:
        {%- if len(case.tuple_case.type) == 1 %}
        return XdrSCVal.forVec([
          XdrSCVal.forSymbol(kind.value),
          {{ to_scval(case.tuple_case.type[0], snake_to_camel(case.tuple_case.name.decode()) + '!') }},
        ]);
        {%- else %}
        final tuple = {{ snake_to_camel(case.tuple_case.name.decode()) }}!;
        return XdrSCVal.forVec([
          XdrSCVal.forSymbol(kind.value),
          {%- for t in case.tuple_case.type %}
          {{ to_scval(t, 'tuple.$' + (loop.index)|string) }}{% if not loop.last %},{% endif %}
          {%- endfor %}
        ]);
        {%- endif %}
      {%- endif %}
      {%- endfor %}
    }
  }

  factory {{ type_name }}.fromScVal(XdrSCVal val) {
    final vec = val.vec!;
    final kind = {{ type_name }}Kind.fromValue(vec[0].sym!.toString());
    
    switch (kind) {
      {%- for case in entry.cases %}
      {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.void_case.name.decode(), False) }}:
        return {{ type_name }}.{{ snake_to_camel(case.void_case.name.decode()) }}();
      {%- else %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }}:
        {%- if len(case.tuple_case.type) == 1 %}
        return {{ type_name }}.{{ snake_to_camel(case.tuple_case.name.decode()) }}(
          {{ from_scval(case.tuple_case.type[0], 'vec[1]') }}
        );
        {%- else %}
        return {{ type_name }}.{{ snake_to_camel(case.tuple_case.name.decode()) }}((
          {%- for i, t in enumerate(case.tuple_case.type) %}
          {{ from_scval(t, 'vec[' + (i + 1)|string + ']') }}{% if not loop.last %},{% endif %}
          {%- endfor %}
        ));
        {%- endif %}
      {%- endif %}
      {%- endfor %}
    }
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;
    if (other is! {{ type_name }}) return false;
    if (kind != other.kind) return false;
    
    switch (kind) {
      {%- for case in entry.cases %}
      {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.void_case.name.decode(), False) }}:
        return true;
      {%- else %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }}:
        return {{ snake_to_camel(case.tuple_case.name.decode()) }} == other.{{ snake_to_camel(case.tuple_case.name.decode()) }};
      {%- endif %}
      {%- endfor %}
    }
  }

  @override
  int get hashCode {
    switch (kind) {
      {%- for case in entry.cases %}
      {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.void_case.name.decode(), False) }}:
        return kind.hashCode;
      {%- else %}
      case {{ type_name }}Kind.{{ snake_to_camel(case.tuple_case.name.decode(), False) }}:
        return Object.hash(kind, {{ snake_to_camel(case.tuple_case.name.decode()) }});
      {%- endif %}
      {%- endfor %}
    }
  }
}
"""
    union_rendered_code = Template(template).render(
        entry=entry,
        type_name=type_name,
        to_dart_type=to_dart_type_bound,
        to_scval=to_scval_bound,
        from_scval=from_scval_bound,
        xdr=xdr,
        len=len,
        snake_to_camel=snake_to_camel,
        enumerate=enumerate,
    )
    return kind_enum_rendered_code + "\n" + union_rendered_code


def render_client(entries: List[xdr.SCSpecFunctionV0], class_name: str):
    template = """
/// Client for interacting with the {{ class_name }} contract
class {{ class_name }} {
  /// The underlying SorobanClient instance
  final SorobanClient _client;

  /// Creates a new {{ class_name }} for the given contract ID
  static Future<{{ class_name }}> forContractId({
    required KeyPair sourceAccountKeyPair,
    required String contractId,
    required Network network,
    required String rpcUrl,
    bool enableServerLogging = false,
  }) async {
    final options = ClientOptions(
      sourceAccountKeyPair: sourceAccountKeyPair,
      contractId: contractId,
      network: network,
      rpcUrl: rpcUrl,
      enableServerLogging: enableServerLogging,
    );
    
    final client = await SorobanClient.forClientOptions(options: options);
    return {{ class_name }}._(client);
  }
  
  /// Private constructor that wraps a SorobanClient
  {{ class_name }}._(this._client);

  /// Gets the contract ID
  String getContractId() => _client.getContractId();

  /// Gets the client options
  ClientOptions getOptions() => _client.getOptions();

  /// Gets the contract specification
  ContractSpec getContractSpec() => _client.getContractSpec();
      
  {%- for entry in entries %}
  
  /// {{ entry.doc.decode() if entry.doc else 'Invokes the ' + entry.name.sc_symbol.decode() + ' method' }}
  {%- if parse_result_type(entry.outputs) == 'void' %}
  Future<void> {{ snake_to_camel(entry.name.sc_symbol.decode()) }}({
  {%- else %}
  Future<{{ parse_result_type(entry.outputs) }}> {{ snake_to_camel(entry.name.sc_symbol.decode()) }}({
  {%- endif %}
    {%- for param in entry.inputs %}
    required {{ to_dart_type(param.type) }} {{ snake_to_camel(param.name.decode()) }},
    {%- endfor %}
    KeyPair? signer,
    int baseFee = 100,
    int transactionTimeout = 300,
    int submitTimeout = 30,
    bool simulate = true,
    bool restore = true,
    bool force = false,
  }) async {
    final List<XdrSCVal> args = [
      {%- for param in entry.inputs %}
      {{ to_scval(param.type, snake_to_camel(param.name.decode())) }},
      {%- endfor %}
    ];
    
    final methodOptions = MethodOptions();
    // You can customize method options here if needed
    
    final result = await _client.invokeMethod(
      name: '{{ entry.name.sc_symbol_r.decode() if entry.name.sc_symbol_r else entry.name.sc_symbol.decode() }}',
      args: args,
      force: force,
      methodOptions: methodOptions,
    );
    
    {%- if parse_result_type(entry.outputs) != 'void' %}
    return {{ parse_result_from_scval(entry.outputs, 'result') }};
    {%- endif %}
  }
  
  /// Builds an AssembledTransaction for the {{ entry.name.sc_symbol.decode() }} method.
  /// This is useful if you need to manipulate the transaction before signing and sending.
  Future<AssembledTransaction> build{{ snake_to_camel(entry.name.sc_symbol.decode(), False) }}Tx({
    {%- for param in entry.inputs %}
    required {{ to_dart_type(param.type) }} {{ snake_to_camel(param.name.decode()) }},
    {%- endfor %}
    MethodOptions? methodOptions,
  }) async {
    final List<XdrSCVal> args = [
      {%- for param in entry.inputs %}
      {{ to_scval(param.type, snake_to_camel(param.name.decode())) }},
      {%- endfor %}
    ];
    
    return await _client.buildInvokeMethodTx(
      name: '{{ entry.name.sc_symbol_r.decode() if entry.name.sc_symbol_r else entry.name.sc_symbol.decode() }}',
      args: args,
      methodOptions: methodOptions,
    );
  }
  {%- endfor %}
}
"""

    # Create wrapper functions with class_name bound
    def to_dart_type_bound(td, nullable=False):
        return to_dart_type(td, nullable, class_name)
    
    def to_scval_bound(td, name):
        return to_scval(td, name, class_name)
    
    def from_scval_bound(td, name):
        return from_scval(td, name, class_name)
    
    def parse_result_type(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return "void"
        elif len(output) == 1:
            return to_dart_type_bound(output[0])
        else:
            types = [to_dart_type_bound(t) for t in output]
            return f"({', '.join(types)})"

    def parse_result_xdr(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return ""
        elif len(output) == 1:
            return from_scval_bound(output[0], "result")
        else:
            # Handle tuple return
            results = []
            for i, t in enumerate(output):
                results.append(from_scval_bound(t, f"result.vec![{i}]"))
            return f"({', '.join(results)})"

    def parse_result_from_scval(output: List[xdr.SCSpecTypeDef], var_name: str):
        if len(output) == 0:
            return ""
        elif len(output) == 1:
            return from_scval_bound(output[0], var_name)
        else:
            # Handle tuple return
            results = []
            for i, t in enumerate(output):
                results.append(from_scval_bound(t, f"{var_name}.vec![{i}]"))
            return f"({', '.join(results)})"

    client_rendered_code = Template(template).render(
        entries=entries,
        to_dart_type=to_dart_type_bound,
        to_scval=to_scval_bound,
        from_scval=from_scval_bound,
        parse_result_type=parse_result_type,
        parse_result_xdr=parse_result_xdr,
        parse_result_from_scval=parse_result_from_scval,
        snake_to_camel=snake_to_camel,
        class_name=class_name,
    )
    return client_rendered_code


def append_underscore(specs: List[xdr.SCSpecEntry]):
    for spec in specs:
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            assert spec.udt_struct_v0 is not None
            if is_keywords(spec.udt_struct_v0.name.decode()):
                spec.udt_struct_v0.name_r = spec.udt_struct_v0.name  # type: ignore[attr-defined]
                spec.udt_struct_v0.name = spec.udt_struct_v0.name + b"_"
            for field in spec.udt_struct_v0.fields:
                if is_keywords(field.name.decode()):
                    field.name_r = field.name  # type: ignore[attr-defined]
                    field.name = field.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            assert spec.udt_union_v0 is not None
            if is_keywords(spec.udt_union_v0.name.decode()):
                spec.udt_union_v0.name_r = spec.udt_union_v0.name  # type: ignore[attr-defined]
                spec.udt_union_v0.name = spec.udt_union_v0.name + b"_"
            for union_case in spec.udt_union_v0.cases:
                if (
                    union_case.kind
                    == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0
                ):
                    if is_keywords(union_case.tuple_case.name.decode()):
                        union_case.tuple_case.name_r = union_case.tuple_case.name  # type: ignore[attr-defined]
                        union_case.tuple_case.name = union_case.tuple_case.name + b"_"
                elif (
                    union_case.kind
                    == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0
                ):
                    if is_keywords(union_case.void_case.name.decode()):
                        union_case.void_case.name_r = union_case.void_case.name  # type: ignore[attr-defined]
                        union_case.void_case.name = union_case.void_case.name + b"_"
                else:
                    raise ValueError(f"Unsupported union case kind: {union_case.kind}")
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0:
            assert spec.function_v0 is not None
            if is_keywords(spec.function_v0.name.sc_symbol.decode()):
                spec.function_v0.name.sc_symbol_r = spec.function_v0.name.sc_symbol  # type: ignore[attr-defined]
                spec.function_v0.name.sc_symbol = spec.function_v0.name.sc_symbol + b"_"
            for param in spec.function_v0.inputs:
                if is_keywords(param.name.decode()):
                    param.name = param.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            assert spec.udt_enum_v0 is not None
            if is_keywords(spec.udt_enum_v0.name.decode()):
                spec.udt_enum_v0.name_r = spec.udt_enum_v0.name  # type: ignore[attr-defined]
                spec.udt_enum_v0.name = spec.udt_enum_v0.name + b"_"
            for enum_case in spec.udt_enum_v0.cases:
                if is_keywords(enum_case.name.decode()):
                    enum_case.name = enum_case.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            assert spec.udt_error_enum_v0 is not None
            if is_keywords(spec.udt_error_enum_v0.name.decode()):
                spec.udt_error_enum_v0.name_r = spec.udt_error_enum_v0.name  # type: ignore[attr-defined]
                spec.udt_error_enum_v0.name = spec.udt_error_enum_v0.name + b"_"
            for error_enum_case in spec.udt_error_enum_v0.cases:
                if is_keywords(error_enum_case.name.decode()):
                    error_enum_case.name = error_enum_case.name + b"_"


def generate_binding(specs: List[xdr.SCSpecEntry], class_name: str) -> str:
    append_underscore(specs)

    generated = []
    generated.append(render_info())
    generated.append(render_imports())

    for spec in specs:
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            generated.append(render_enum(spec.udt_enum_v0, class_name))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            generated.append(render_error_enum(spec.udt_error_enum_v0, class_name))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            if is_tuple_struct(spec.udt_struct_v0):
                generated.append(render_tuple_struct(spec.udt_struct_v0, class_name))
            else:
                generated.append(render_struct(spec.udt_struct_v0, class_name))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            generated.append(render_union(spec.udt_union_v0, class_name))

    function_specs: List[xdr.SCSpecFunctionV0] = [
        spec.function_v0
        for spec in specs
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0
        and not spec.function_v0.name.sc_symbol.decode().startswith("__")
    ]
    generated.append(render_client(function_specs, class_name))
    return "\n".join(generated)


@click.command(name="flutter")
@click.option(
    "--contract-id", required=True, help="The contract ID to generate bindings for"
)
@click.option(
    "--rpc-url", default="https://mainnet.sorobanrpc.com", help="Soroban RPC URL"
)
@click.option(
    "--output",
    default=None,
    help="Output directory for generated bindings, defaults to current directory",
)
@click.option(
    "--class-name",
    default="Contract",
    help="Class name prefix for generated bindings, defaults to 'Contract'",
)
def command(contract_id: str, rpc_url: str, output: str, class_name: str):
    """Generate Flutter/Dart bindings for a Soroban contract"""
    if not StrKey.is_valid_contract(contract_id):
        click.echo(f"Invalid contract ID: {contract_id}", err=True)
        raise click.Abort()

    # Use current directory if output is not specified
    if output is None:
        output = os.getcwd()
    try:
        specs = get_specs_by_contract_id(contract_id, rpc_url)
    except Exception as e:
        click.echo(f"Get contract specs failed: {e}", err=True)
        raise click.Abort()

    click.echo("Generating Flutter bindings")
    generated = generate_binding(specs, class_name=class_name)

    if not os.path.exists(output):
        os.makedirs(output)
    output_path = os.path.join(
        output, f"{camel_to_snake(class_name)}_client.dart"
    )
    with open(output_path, "w") as f:
        f.write(generated)
    click.echo(f"Generated Flutter bindings to {output_path}")


if __name__ == "__main__":
    command()



from stellar_sdk import SorobanServer
from stellar_sdk import xdr, Address

from stellar_contract_bindings.metadata import (
    parse_contract_metadata,
    get_token_sc_spec_entry,
)


def get_specs_by_wasm_bytes(wasm: bytes) -> list[xdr.SCSpecEntry]:
    """Get the contract specs by wasm bytes.

    :param wasm: The wasm bytes.
    :return: The contract specs.
    """
    meta_data = parse_contract_metadata(wasm)
    return meta_data.spec


def get_specs_by_wasm_file(wasm_file_path: str) -> list[xdr.SCSpecEntry]:
    """Get the contract specs by wasm file path.

    :param wasm_file_path: The wasm file path.
    :return: The contract specs.
    """
    with open(wasm_file_path, "rb") as f:
        wasm = f.read()
    return get_specs_by_wasm_bytes(wasm)


def get_specs_by_wasm_hash(wasm_hash: bytes, rpc_url: str) -> list[xdr.SCSpecEntry]:
    """Get the contract wasm by wasm hash.

    :param wasm_hash: The wasm hash.
    :param rpc_url: The Soroban RPC URL.
    :return: The contract wasm.
    :raises ValueError: If wasm not found.
    """
    with SorobanServer(rpc_url) as server:
        key = xdr.LedgerKey(
            xdr.LedgerEntryType.CONTRACT_CODE,
            contract_code=xdr.LedgerKeyContractCode(hash=xdr.Hash(wasm_hash)),
        )
        resp = server.get_ledger_entries([key])
        if not resp.entries:
            raise ValueError(f"Wasm not found, wasm id: {wasm_hash.hex()}")
        data = xdr.LedgerEntryData.from_xdr(resp.entries[0].xdr)
        meta_data = data.contract_code.code
        return get_specs_by_wasm_bytes(meta_data)


def get_specs_by_contract_id(contract_id: str, rpc_url: str) -> list[xdr.SCSpecEntry]:
    """Get the wasm hash by contract id.

    :param contract_id: The contract id.
    :param rpc_url: The Soroban RPC URL.
    :return: The wasm hash.
    :raises ValueError: If contract not found.
    """
    with SorobanServer(rpc_url) as server:
        key = xdr.LedgerKey(
            xdr.LedgerEntryType.CONTRACT_DATA,
            contract_data=xdr.LedgerKeyContractData(
                contract=Address(contract_id).to_xdr_sc_address(),
                key=xdr.SCVal(xdr.SCValType.SCV_LEDGER_KEY_CONTRACT_INSTANCE),
                durability=xdr.ContractDataDurability.PERSISTENT,
            ),
        )
        resp = server.get_ledger_entries([key])
        if not resp.entries:
            raise ValueError(f"Contract not found, contract id: {contract_id}")
        data = xdr.LedgerEntryData.from_xdr(resp.entries[0].xdr)
        if (
                data.contract_data.val.instance.executable.type
                == xdr.ContractExecutableType.CONTRACT_EXECUTABLE_STELLAR_ASSET
        ):
            return get_token_sc_spec_entry()
        elif (
                data.contract_data.val.instance.executable.type
                == xdr.ContractExecutableType.CONTRACT_EXECUTABLE_WASM
        ):
            return get_specs_by_wasm_hash(
                data.contract_data.val.instance.executable.wasm_hash.hash, rpc_url
            )
        else:
            raise ValueError(
                f"Unknown executable type, type: {data.contract_data.val.instance.executable.type}"
            )


if __name__ == "__main__":
    specs = get_specs_by_contract_id(
        "CAS3J7GYLGXMF6TDJBBYYSE3HQ6BBSMLNUQ34T6TZMYMW2EVH34XOWMA",
        "https://mainnet.sorobanrpc.com",
    )
    for spec in specs:
        print(spec)



import keyword
import os
from typing import List
import black

import click
from jinja2 import Template
from stellar_sdk import __version__ as stellar_sdk_version, StrKey
from stellar_sdk import xdr

from stellar_contract_bindings import __version__ as stellar_contract_bindings_version
from stellar_contract_bindings.utils import get_specs_by_contract_id


def is_tuple_struct(entry: xdr.SCSpecUDTStructV0) -> bool:
    # ex. <SCSpecUDTStructV0 [doc=b'', lib=b'', name=b'TupleStruct', fields=[<SCSpecUDTStructFieldV0 [doc=b'', name=b'0', type=<SCSpecTy...>, <SCSpecUDTStructFieldV0 [doc=b'', name=b'1', type=<SCSpecTypeDef [type=2000, udt=<SCSpecTypeUDT [name=b'SimpleEnum']>]>]>]]>
    return all(f.name.isdigit() for f in entry.fields)


def camel_to_snake(text: str) -> str:
    result = text[0].lower()
    for char in text[1:]:
        if char.isupper():
            result += "_" + char.lower()
        else:
            result += char
    return result


def to_py_type(td: xdr.SCSpecTypeDef, input_type: bool = False):
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return "xdr.SCVal"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return "bool"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return "None"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return "int"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return "bytes"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return "bytes"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return "str"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS or t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return "Union[Address, str]" if input_type else "Address"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        raise NotImplementedError("SC_SPEC_TYPE_MUXED_ADDRESS is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return f"Optional[{to_py_type(td.option.value_type, input_type)}]"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return to_py_type(ok_t, input_type)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return f"List[{to_py_type(td.vec.element_type, input_type)}]"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"Dict[{to_py_type(td.map.key_type, input_type)}, {to_py_type(td.map.value_type, input_type)}]"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            # () equivalent to None in Python
            return "None"
        types = [to_py_type(t, input_type) for t in td.tuple.value_types]
        return f"Tuple[{', '.join(types)}]"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"bytes"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        return td.udt.name.decode()
    raise ValueError(f"Unsupported SCValType: {t}")


def to_scval(td: xdr.SCSpecTypeDef, name: str):
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"{name}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"scval.to_bool({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"scval.to_void()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"scval.to_uint32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"scval.to_int32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"scval.to_uint64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"scval.to_int64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"scval.to_timepoint({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"scval.to_duration({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"scval.to_uint128({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"scval.to_int128({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"scval.to_uint256({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"scval.to_int256({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"scval.to_bytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"scval.to_string({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"scval.to_symbol({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS or t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"scval.to_address({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        raise NotImplementedError("SC_SPEC_TYPE_MUXED_ADDRESS is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return f"{to_scval(td.option.value_type, name)} if {name} is not None else scval.to_void()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        return NotImplementedError("SC_SPEC_TYPE_RESULT is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return f"scval.to_vec([{to_scval(td.vec.element_type, 'e')} for e in {name}])"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"scval.to_map({{{to_scval(td.map.key_type, 'k')}: {to_scval(td.map.value_type, 'v')} for k, v in {name}.items()}})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        types = [
            to_scval(t, f"{name}[{i}]") for i, t in enumerate(td.tuple.value_types)
        ]
        return f"scval.to_tuple_struct([{', '.join(types)}])"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"scval.to_bytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        return f"{name}.to_scval()"
    raise ValueError(f"Unsupported SCValType: {t}")


def from_scval(td: xdr.SCSpecTypeDef, name: str):
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"{name}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"scval.from_bool({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"scval.from_void()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"scval.from_uint32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"scval.from_int32({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"scval.from_uint64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"scval.from_int64({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"scval.from_timepoint({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"scval.from_duration({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"scval.from_uint128({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"scval.from_int128({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"scval.from_uint256({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"scval.from_int256({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"scval.from_bytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"scval.from_string({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"scval.from_symbol({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS or t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"scval.from_address({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        raise NotImplementedError("SC_SPEC_TYPE_MUXED_ADDRESS is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return f"{from_scval(td.option.value_type, name)} if {name}.type != xdr.SCValType.SCV_VOID else scval.from_void({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return f"{from_scval(ok_t, name)}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        return (
            f"[{from_scval(td.vec.element_type, 'e')} for e in scval.from_vec({name})]"
        )
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"{{{from_scval(td.map.key_type, 'k')}: {from_scval(td.map.value_type, 'v')} for k, v in scval.from_map({name}).items()}}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "None"
        elements = f"scval.from_tuple_struct({name})"
        types = [
            from_scval(t, f"{elements}[{i}]")
            for i, t in enumerate(td.tuple.value_types)
        ]
        return f"({', '.join(types)})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"scval.from_bytes({name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        return f"{td.udt.name.decode()}.from_scval({name})"
    raise NotImplementedError(f"Unsupported SCValType: {t}")


def render_info():
    return f"# This file was generated by stellar_contract_bindings v{stellar_contract_bindings_version} and stellar_sdk v{stellar_sdk_version}."


def render_imports(client_type: str = "both"):
    template = """
from __future__ import annotations

from enum import IntEnum, Enum
from typing import Dict, List, Tuple, Optional, Union

from stellar_sdk import scval, xdr, Address, MuxedAccount, Keypair
{%- if client_type == "sync" or client_type == "both" %}
from stellar_sdk.contract import AssembledTransaction, ContractClient
{%- endif %}
{%- if client_type == "async" or client_type == "both" %}
from stellar_sdk.contract import AssembledTransactionAsync, ContractClientAsync
{%- endif %}

NULL_ACCOUNT = "GAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAWHF"
"""
    rendered_code = Template(template).render(client_type=client_type)
    return rendered_code


def render_enum(entry: xdr.SCSpecUDTEnumV0):
    template = """
class {{ entry.name.decode() }}(IntEnum):
    {%- if entry.doc %}
    '''{{ entry.doc.decode() }}'''
    {%- endif %}
    {%- for case in entry.cases %}
    {{ case.name.decode() }} = {{ case.value.uint32 }}
    {%- endfor %}
    def to_scval(self) -> xdr.SCVal:
        return scval.to_uint32(self.value)

    @classmethod
    def from_scval(cls, val: xdr.SCVal):
        return cls(scval.from_uint32(val))
"""

    rendered_code = Template(template).render(entry=entry)
    return rendered_code


def render_error_enum(entry: xdr.SCSpecUDTErrorEnumV0):
    template = """
class {{ entry.name.decode() }}(IntEnum):
    {%- if entry.doc %}
    '''{{ entry.doc.decode() }}'''
    {%- endif %}
    {%- for case in entry.cases %}
    {{ case.name.decode() }} = {{ case.value.uint32 }}
    {%- endfor %}
    def to_scval(self) -> xdr.SCVal:
        return scval.to_uint32(self.value)

    @classmethod
    def from_scval(cls, val: xdr.SCVal):
        return cls(scval.from_uint32(val))
    """

    rendered_code = Template(template).render(entry=entry)
    return rendered_code


def render_struct(entry: xdr.SCSpecUDTStructV0):
    template = """
class {{ entry.name.decode() }}:
    {%- if entry.doc %}
    '''{{ entry.doc.decode() }}'''
    {%- endif %}
    {%- for field in entry.fields %}
    {{ field.name.decode() }}: {{ to_py_type(field.type) }}
    {%- endfor %}

    def __init__(self, {% for field in entry.fields %}{{ field.name.decode() }}: {{ to_py_type(field.type, True) }}{% if not loop.last %}, {% endif %}{% endfor %}):
        {%- for field in entry.fields %}
        self.{{ field.name.decode() }} = {{ field.name.decode() }}
        {%- endfor %}

    def to_scval(self) -> xdr.SCVal:
        return scval.to_struct({
            {%- for field in entry.fields %}
            '{{ field.name_r.decode() if field.name_r else field.name.decode() }}': {{ to_scval(field.type, 'self.' ~ field.name.decode()) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        })

    @classmethod
    def from_scval(cls, val: xdr.SCVal):
        elements = scval.from_struct(val)
        return cls(
            {%- for index, field in enumerate(entry.fields) %}
            {{ from_scval(field.type, 'elements["' ~ (field.name_r.decode() if field.name_r else field.name.decode()) ~ '"]') }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, {{ entry.name.decode() }}):
            return NotImplemented
        return {% for field in entry.fields %}self.{{ field.name.decode() }} == other.{{ field.name.decode() }}{% if not loop.last %} and {% endif %}{% endfor %}

    def __hash__(self) -> int:
        return hash(({% for field in entry.fields %}self.{{ field.name.decode() }}{% if not loop.last %}, {% endif %}{% endfor %}))
"""

    rendered_code = Template(template).render(
        entry=entry,
        to_py_type=to_py_type,
        to_scval=to_scval,
        from_scval=from_scval,
        enumerate=enumerate,
    )
    return rendered_code


def render_tuple_struct(entry: xdr.SCSpecUDTStructV0):
    template = """
class {{ entry.name.decode() }}:
    {%- if entry.doc %}
    '''{{ entry.doc.decode() }}'''
    {%- endif %}

    def __init__(self, value: Tuple[{% for f in entry.fields %}{{ to_py_type(f.type, True) }}{% if not loop.last %}, {% endif %}{% endfor %}]):
        self.value = value

    def to_scval(self) -> xdr.SCVal:
        return scval.to_tuple_struct([{% for f in entry.fields %}{{ to_scval(f.type, 'self.value[' ~ f.name.decode() ~ ']') }}{% if not loop.last %}, {% endif %}{% endfor %}]) 
    
    @classmethod
    def from_scval(cls, val: xdr.SCVal):
        elements = scval.from_tuple_struct(val)
        values = ({% for f in entry.fields %}{{ from_scval(f.type, 'elements[' ~ f.name.decode() ~ ']') }}, {% endfor %})
        return cls(values)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, {{ entry.name.decode() }}):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)
"""

    rendered_code = Template(template).render(
        entry=entry, to_py_type=to_py_type, to_scval=to_scval, from_scval=from_scval
    )
    return rendered_code


def render_union(entry: xdr.SCSpecUDTUnionV0):
    kind_enum_template = """
class {{ entry.name.decode() }}Kind(Enum):
    {%- for case in entry.cases %}
    {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
    {{ case.void_case.name.decode() }} = '{{ case.void_case.name_r.decode() if case.void_case.name_r else case.void_case.name.decode() }}'
    {%- else %}
    {{ case.tuple_case.name.decode() }} = '{{ case.tuple_case.name.decode() if case.tuple_case.name_r else case.tuple_case.name.decode() }}'
    {%- endif %}
    {%- endfor %}
"""

    kind_enum_rendered_code = Template(kind_enum_template).render(entry=entry, xdr=xdr)

    template = """
class {{ entry.name.decode() }}:
    {%- if entry.doc %}
    '''{{ entry.doc.decode() }}'''
    {%- endif %}
    def __init__(self,
        kind: {{ entry.name.decode() }}Kind,
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
        {%- if len(case.tuple_case.type) == 1 %}
        {{ camel_to_snake(case.tuple_case.name.decode()) }}: Optional[{{ to_py_type(case.tuple_case.type[0], True) }}] = None,
        {%- else %}
        {{ camel_to_snake(case.tuple_case.name.decode()) }}: Optional[Tuple[{% for f in case.tuple_case.type %}{{ to_py_type(f, True) }}{% if not loop.last %}, {% endif %}{% endfor %}]] = None,
        {%- endif %}
        {%- endif %}
        {%- endfor %}
    ):
        self.kind = kind
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %} 
        self.{{ camel_to_snake(case.tuple_case.name.decode()) }} = {{ camel_to_snake(case.tuple_case.name.decode()) }}
        {%- endif %}
        {%- endfor %}

    def to_scval(self) -> xdr.SCVal:
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
        if self.kind == {{ entry.name.decode() }}Kind.{{ case.void_case.name.decode() }}:
            return scval.to_enum(self.kind.name, None)
        {%- else %}
        if self.kind == {{ entry.name.decode() }}Kind.{{ case.tuple_case.name.decode() }}:
        {%- if len(case.tuple_case.type) == 1 %}
            assert self.{{ camel_to_snake(case.tuple_case.name.decode()) }} is not None
            return scval.to_enum(self.kind.name, {{ to_scval(case.tuple_case.type[0], 'self.' ~ camel_to_snake(case.tuple_case.name.decode())) }})
        {%- else %}
            assert isinstance(self.{{ camel_to_snake(case.tuple_case.name.decode()) }}, tuple)
            return scval.to_enum(self.kind.name, [
                {%- for t in case.tuple_case.type %}
                {{ to_scval(t, 'self.' + camel_to_snake(case.tuple_case.name.decode()) + '[' + loop.index0|string + ']') }}{% if not loop.last %},{% endif %}
                {%- endfor %}
            ])
        {%- endif %}
        {%- endif %}
        {%- endfor %}
        raise ValueError(f"Invalid kind: {self.kind}")
    
    @classmethod
    def from_scval(cls, val: xdr.SCVal):
        elements = scval.from_enum(val)
        kind = {{ entry.name.decode() }}Kind(elements[0])
        
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
        if kind == {{ entry.name.decode() }}Kind.{{ case.void_case.name.decode() }}:
            return cls(kind)
        {%- else %}
        if kind == {{ entry.name.decode() }}Kind.{{ case.tuple_case.name.decode() }}:
        {%- if len(case.tuple_case.type) == 1 %}
            assert elements[1] is not None and isinstance(elements[1], xdr.SCVal)
            return cls(kind, {{ camel_to_snake(case.tuple_case.name.decode()) }}={{ from_scval(case.tuple_case.type[0], 'elements[1]') }})
        {%- else %}
            assert elements[1] is not None and isinstance(elements[1], list)
            return cls(kind, {{ camel_to_snake(case.tuple_case.name.decode()) }}=(
                {%- for i, t in enumerate(case.tuple_case.type) %}
                {{ from_scval(t, 'elements[1][' + loop.index0|string + ']') }}{% if not loop.last %},{% endif %}
                {%- endfor %}
            ))
        {%- endif %}
        {%- endif %}
        {%- endfor %}
        raise ValueError(f"Invalid kind: {kind}")
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, {{ entry.name.decode() }}):
            return NotImplemented
        if self.kind != other.kind:
            return False
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
        if self.kind == {{ entry.name.decode() }}Kind.{{ case.tuple_case.name.decode() }}:
            return self.{{ camel_to_snake(case.tuple_case.name.decode()) }} == other.{{ camel_to_snake(case.tuple_case.name.decode()) }}
        {%- endif %}
        {%- endfor %}
        return True

    def __hash__(self) -> int:
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
        if self.kind == {{ entry.name.decode() }}Kind.{{ case.tuple_case.name.decode() }}:
            return hash((self.kind, self.{{ camel_to_snake(case.tuple_case.name.decode()) }}))
        {%- endif %}
        {%- endfor %}
        return hash(self.kind)
"""
    union_rendered_code = Template(template).render(
        entry=entry,
        to_py_type=to_py_type,
        to_scval=to_scval,
        from_scval=from_scval,
        xdr=xdr,
        len=len,
        camel_to_snake=camel_to_snake,
        enumerate=enumerate,
    )
    return kind_enum_rendered_code + "\n" + union_rendered_code


def render_client(entries: List[xdr.SCSpecFunctionV0], client_type: str):
    template = '''
{%- if client_type == "sync" or client_type == "both" %}
class Client(ContractClient):
    {%- for entry in entries %}
    def {{ entry.name.sc_symbol.decode() }}(self, {% for param in entry.inputs %}{{ param.name.decode() }}: {{ to_py_type(param.type, True) }}, {% endfor %} source: Union[str, MuxedAccount] = NULL_ACCOUNT, signer: Optional[Keypair] = None, base_fee: int = 100, transaction_timeout: int = 300, submit_timeout: int = 30, simulate: bool = True, restore: bool = True) -> AssembledTransaction[{{ parse_result_type(entry.outputs) }}]:
        {%- if entry.doc %}
        """{{ entry.doc.decode() }}"""
        {%- endif %}
        return self.invoke('{{ entry.name.sc_symbol_r.decode() if entry.name.sc_symbol_r else entry.name.sc_symbol.decode() }}', [{% for param in entry.inputs %}{{ to_scval(param.type, param.name.decode()) }}{% if not loop.last %}, {% endif %}{% endfor %}], parse_result_xdr_fn={{ parse_result_xdr_fn(entry.outputs) }}, source = source, signer = signer, base_fee = base_fee, transaction_timeout = transaction_timeout, submit_timeout = submit_timeout, simulate = simulate, restore = restore)
    {%- endfor %}
{%- endif %}

{%- if client_type == "async" or client_type == "both" %}
class ClientAsync(ContractClientAsync):
    {%- for entry in entries %}
    async def {{ entry.name.sc_symbol.decode() }}(self, {% for param in entry.inputs %}{{ param.name.decode() }}: {{ to_py_type(param.type, True) }}, {% endfor %} source: Union[str, MuxedAccount] = NULL_ACCOUNT, signer: Optional[Keypair] = None, base_fee: int = 100, transaction_timeout: int = 300, submit_timeout: int = 30, simulate: bool = True, restore: bool = True) -> AssembledTransactionAsync[{{ parse_result_type(entry.outputs) }}]:
        {%- if entry.doc %}
        """{{ entry.doc.decode() }}"""
        {%- endif %}
        return await self.invoke('{{ entry.name.sc_symbol_r.decode() if entry.name.sc_symbol_r else entry.name.sc_symbol.decode() }}', [{% for param in entry.inputs %}{{ to_scval(param.type, param.name.decode()) }}{% if not loop.last %}, {% endif %}{% endfor %}], parse_result_xdr_fn={{ parse_result_xdr_fn(entry.outputs) }}, source = source, signer = signer, base_fee = base_fee, transaction_timeout = transaction_timeout, submit_timeout = submit_timeout, simulate = simulate, restore = restore)
    {%- endfor %}
{%- endif %}
'''

    def parse_result_type(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return "None"
        elif len(output) == 1:
            return to_py_type(output[0])
        else:
            return f"Tuple[{', '.join([to_py_type(t) for t in output])}]"

    def parse_result_xdr_fn(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return "lambda _: None"
        elif len(output) == 1:
            return f'lambda v: {from_scval(output[0], "v")}'
        else:
            raise NotImplementedError(
                "Tuple return type is not supported, please report this issue"
            )

    client_rendered_code = Template(template).render(
        entries=entries,
        to_py_type=to_py_type,
        to_scval=to_scval,
        parse_result_type=parse_result_type,
        parse_result_xdr_fn=parse_result_xdr_fn,
        client_type=client_type,
    )
    return client_rendered_code


# append _ to keyword
def append_underscore(specs: List[xdr.SCSpecEntry]):
    for spec in specs:
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            assert spec.udt_struct_v0 is not None
            if keyword.iskeyword(spec.udt_struct_v0.name.decode()):
                spec.udt_struct_v0.name_r = spec.udt_struct_v0.name  # type: ignore[attr-defined]
                spec.udt_struct_v0.name = spec.udt_struct_v0.name + b"_"
            for field in spec.udt_struct_v0.fields:
                if keyword.iskeyword(field.name.decode()):
                    field.name_r = field.name  # type: ignore[attr-defined]
                    field.name = field.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            assert spec.udt_union_v0 is not None
            if keyword.iskeyword(spec.udt_union_v0.name.decode()):
                spec.udt_union_v0.name_r = spec.udt_union_v0.name  # type: ignore[attr-defined]
                spec.udt_union_v0.name = spec.udt_union_v0.name + b"_"
            for union_case in spec.udt_union_v0.cases:
                if (
                    union_case.kind
                    == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0
                ):
                    if keyword.iskeyword(union_case.tuple_case.name.decode()):
                        union_case.tuple_case.name_r = union_case.tuple_case.name  # type: ignore[attr-defined]
                        union_case.tuple_case.name = union_case.tuple_case.name + b"_"
                elif (
                    union_case.kind
                    == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0
                ):
                    if keyword.iskeyword(union_case.void_case.name.decode()):
                        union_case.void_case.name_r = union_case.void_case.name  # type: ignore[attr-defined]
                        union_case.void_case.name = union_case.void_case.name + b"_"
                else:
                    raise ValueError(f"Unsupported union case kind: {union_case.kind}")
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0:
            assert spec.function_v0 is not None
            if keyword.iskeyword(spec.function_v0.name.sc_symbol.decode()):
                spec.function_v0.name.sc_symbol_r = spec.function_v0.name.sc_symbol  # type: ignore[attr-defined]
                spec.function_v0.name.sc_symbol = spec.function_v0.name.sc_symbol + b"_"
            for param in spec.function_v0.inputs:
                if keyword.iskeyword(param.name.decode()):
                    param.name = param.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            assert spec.udt_enum_v0 is not None
            if keyword.iskeyword(spec.udt_enum_v0.name.decode()):
                spec.udt_enum_v0.name_r = spec.udt_enum_v0.name  # type: ignore[attr-defined]
                spec.udt_enum_v0.name = spec.udt_enum_v0.name + b"_"
            for enum_case in spec.udt_enum_v0.cases:
                if keyword.iskeyword(enum_case.name.decode()):
                    enum_case.name = enum_case.name + b"_"
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            assert spec.udt_error_enum_v0 is not None
            if keyword.iskeyword(spec.udt_error_enum_v0.name.decode()):
                spec.udt_error_enum_v0.name_r = spec.udt_error_enum_v0.name  # type: ignore[attr-defined]
                spec.udt_error_enum_v0.name = spec.udt_error_enum_v0.name + b"_"
            for error_enum_case in spec.udt_error_enum_v0.cases:
                if keyword.iskeyword(error_enum_case.name.decode()):
                    error_enum_case.name = error_enum_case.name + b"_"


def generate_binding(specs: List[xdr.SCSpecEntry], client_type: str) -> str:
    append_underscore(specs)

    generated = []
    generated.append(render_info())
    generated.append(render_imports(client_type))

    for spec in specs:
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            generated.append(render_enum(spec.udt_enum_v0))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            generated.append(render_error_enum(spec.udt_error_enum_v0))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            if is_tuple_struct(spec.udt_struct_v0):
                generated.append(render_tuple_struct(spec.udt_struct_v0))
            else:
                generated.append(render_struct(spec.udt_struct_v0))
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            generated.append(render_union(spec.udt_union_v0))

    function_specs: List[xdr.SCSpecFunctionV0] = [
        spec.function_v0
        for spec in specs
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0
        and not spec.function_v0.name.sc_symbol.decode().startswith("__")
    ]
    generated.append(render_client(function_specs, client_type))
    return "\n".join(generated)


@click.command(name="python")
@click.option(
    "--contract-id", required=True, help="The contract ID to generate bindings for"
)
@click.option(
    "--rpc-url", default="https://mainnet.sorobanrpc.com", help="Soroban RPC URL"
)
@click.option(
    "--output",
    default=None,
    help="Output directory for generated bindings, defaults to current directory",
)
@click.option(
    "--client-type",
    type=click.Choice(["sync", "async", "both"], case_sensitive=False),
    default="both",
    help="Client type to generate, defaults to both sync and async",
)
def command(contract_id: str, rpc_url: str, output: str, client_type: str):
    """Generate Python bindings for a Soroban contract"""
    if not StrKey.is_valid_contract(contract_id):
        click.echo(f"Invalid contract ID: {contract_id}", err=True)
        raise click.Abort()

    # Use current directory if output is not specified
    if output is None:
        output = os.getcwd()
    try:
        specs = get_specs_by_contract_id(contract_id, rpc_url)
    except Exception as e:
        click.echo(f"Get contract specs failed: {e}", err=True)
        raise click.Abort()

    click.echo("Generating Python bindings")
    generated = generate_binding(specs, client_type=client_type)
    try:
        generated = black.format_str(generated, mode=black.Mode())
    except Exception as e:
        click.echo(
            f"formatting failed, there may be issues with the generated binding, please report to us: {e}",
            err=True,
        )

    if not os.path.exists(output):
        os.makedirs(output)
    output_path = os.path.join(output, "bindings.py")
    with open(output_path, "w") as f:
        f.write(generated)
    click.echo(f"Generated Python bindings to {output_path}")


if __name__ == "__main__":
    from stellar_contract_bindings.metadata import parse_contract_metadata

    wasm_file = "/Users/overcat/repo/lightsail/stellar-contract-bindings/tests/contracts/target/wasm32v1-none/release/python.wasm"
    with open(wasm_file, "rb") as f:
        wasm = f.read()

    specs = parse_contract_metadata(wasm).spec
    generated = generate_binding(specs, client_type="both")
    print(black.format_str(generated, mode=black.Mode()))



import os
from typing import List

import click
from jinja2 import Template
from stellar_sdk import __version__ as stellar_sdk_version, StrKey
from stellar_sdk import xdr

from stellar_contract_bindings import __version__ as stellar_contract_bindings_version
from stellar_contract_bindings.utils import get_specs_by_contract_id


def is_php_keyword(word: str) -> bool:
    """Check if a word is a PHP reserved keyword."""
    return word.lower() in [
        "abstract", "and", "array", "as", "break", "callable", "case", "catch",
        "class", "clone", "const", "continue", "declare", "default", "die", "do",
        "echo", "else", "elseif", "empty", "enddeclare", "endfor", "endforeach",
        "endif", "endswitch", "endwhile", "eval", "exit", "extends", "final",
        "finally", "fn", "for", "foreach", "function", "global", "goto", "if",
        "implements", "include", "include_once", "instanceof", "insteadof",
        "interface", "isset", "list", "match", "namespace", "new", "or", "print",
        "private", "protected", "public", "readonly", "require", "require_once",
        "return", "static", "switch", "throw", "trait", "try", "unset", "use",
        "var", "while", "xor", "yield", "yield_from",
        "__halt_compiler", "__class__", "__dir__", "__file__", "__function__",
        "__line__", "__method__", "__namespace__", "__trait__",
        "int", "float", "bool", "string", "true", "false", "null", "void",
        "iterable", "object", "resource", "mixed", "never"
    ]


def is_tuple_struct(entry: xdr.SCSpecUDTStructV0) -> bool:
    """Check if a struct is a tuple struct (fields are numeric)."""
    return all(f.name.isdigit() for f in entry.fields)


def snake_to_pascal(text: str) -> str:
    """Convert snake_case to PascalCase."""
    parts = text.split("_")
    return "".join(part.capitalize() for part in parts)


def snake_to_camel(text: str) -> str:
    """Convert snake_case to camelCase."""
    parts = text.split("_")
    return parts[0].lower() + "".join(part.capitalize() for part in parts[1:])


def camel_to_snake(text: str) -> str:
    """Convert CamelCase to snake_case."""
    result = text[0].lower()
    for char in text[1:]:
        if char.isupper():
            result += "_" + char.lower()
        else:
            result += char
    return result


def escape_keyword(name: str, context: str = "property") -> str:
    """Escape PHP keywords by appending underscore."""
    if is_php_keyword(name):
        return f"{name}_"
    return name


def prefixed_type_name(type_name: str, class_name: str) -> str:
    """Prefix a type name with the class name to avoid conflicts.
    
    Args:
        type_name: The original type name from the contract spec
        class_name: The class name to use as prefix
    
    Returns:
        The prefixed type name (e.g., "DataKey" -> "TokenContractDataKey")
    """
    # Don't prefix primitive types or SDK types
    if type_name in ['string', 'bool', 'int', 'float', 'array', 'Address', 
                     'XdrSCVal', 'XdrSCMapEntry', 'XdrSCValType']:
        return type_name
    return f"{class_name}{type_name}"


def to_php_type(td: xdr.SCSpecTypeDef, nullable: bool = False, class_name: str = "") -> str:
    """Convert Soroban type to PHP type hint."""
    t = td.type
    nullable_prefix = "?" if nullable else ""
    
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"{nullable_prefix}XdrSCVal"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"{nullable_prefix}bool"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return "void"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t in [xdr.SCSpecType.SC_SPEC_TYPE_U32, xdr.SCSpecType.SC_SPEC_TYPE_I32,
             xdr.SCSpecType.SC_SPEC_TYPE_U64, xdr.SCSpecType.SC_SPEC_TYPE_I64,
             xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT, xdr.SCSpecType.SC_SPEC_TYPE_DURATION]:
        return f"{nullable_prefix}int"
    if t in [xdr.SCSpecType.SC_SPEC_TYPE_U128, xdr.SCSpecType.SC_SPEC_TYPE_I128,
             xdr.SCSpecType.SC_SPEC_TYPE_U256, xdr.SCSpecType.SC_SPEC_TYPE_I256]:
        return f"{nullable_prefix}string" 
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"{nullable_prefix}string"  # PHP uses string for bytes
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"{nullable_prefix}string"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"{nullable_prefix}string"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"{nullable_prefix}Address"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"{nullable_prefix}Address"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        return to_php_type(td.option.value_type, nullable=True, class_name=class_name)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return to_php_type(ok_t, nullable, class_name=class_name)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        inner_type = to_php_type(td.vec.element_type, class_name=class_name)
        return f"{nullable_prefix}array"  # PHP arrays
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        return f"{nullable_prefix}array"  # PHP associative arrays
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "void"
        return f"{nullable_prefix}array"  # PHP arrays for tuples
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"{nullable_prefix}string"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        udt_name = td.udt.name.decode()
        return f"{nullable_prefix}{prefixed_type_name(udt_name, class_name)}"
    raise ValueError(f"Unsupported SCValType: {t}")


def to_scval(td: xdr.SCSpecTypeDef, name: str, class_name: str = "") -> str:
    """Generate PHP code to convert a value to XdrSCVal."""
    t = td.type
    # Add $ prefix only if name doesn't already start with $ or this->
    if name.startswith('$') or name.startswith('this->'):
        var_name = name
    else:
        var_name = f"${name}"
    
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return var_name
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"XdrSCVal::forBool({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"XdrSCVal::forVoid()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"XdrSCVal::forU32({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"XdrSCVal::forI32({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"XdrSCVal::forU64({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"XdrSCVal::forI64({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"XdrSCVal::forTimepoint({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"XdrSCVal::forDuration({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U128:
        return f"XdrSCVal::forU128BigInt({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I128:
        return f"XdrSCVal::forI128BigInt({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U256:
        return f"XdrSCVal::forU256BigInt({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I256:
        return f"XdrSCVal::forI256BigInt({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"XdrSCVal::forBytes({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"XdrSCVal::forString({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"XdrSCVal::forSymbol({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"{var_name}->toXdrSCVal()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"{var_name}->toXdrSCVal()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        inner = to_scval(td.option.value_type, name, class_name)
        return f"({var_name} !== null ? {inner} : XdrSCVal::forVoid())"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        return NotImplementedError("SC_SPEC_TYPE_RESULT is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        element_conversion = to_scval(td.vec.element_type, "item", class_name)
        return f"XdrSCVal::forVec(array_map(fn($item) => {element_conversion}, {var_name}))"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        key_conv = to_scval(td.map.key_type, "k", class_name)
        val_conv = to_scval(td.map.value_type, "v", class_name)
        return f"XdrSCVal::forMap(array_map(fn($k, $v) => [{key_conv}, {val_conv}], array_keys({var_name}), {var_name}))"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "XdrSCVal::forVoid()"
        conversions = [to_scval(td.tuple.value_types[i], f"{name}[{i}]", class_name) for i in range(len(td.tuple.value_types))]
        return f"XdrSCVal::forTupleStruct([{', '.join(conversions)}])"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"XdrSCVal::forBytes({var_name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        return f"{var_name}->toSCVal()"
    raise ValueError(f"Unsupported SCValType: {t}")


def from_scval(td: xdr.SCSpecTypeDef, name: str, class_name: str = "") -> str:
    """Generate PHP code to convert from XdrSCVal to a PHP value."""
    t = td.type
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VAL:
        return f"${name}"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BOOL:
        return f"${name}->b"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VOID:
        return f"null"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ERROR:
        raise NotImplementedError("SC_SPEC_TYPE_ERROR is not supported")
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U32:
        return f"${name}->u32"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I32:
        return f"${name}->i32"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_U64:
        return f"${name}->u64"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_I64:
        return f"${name}->i64"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
        return f"${name}->timepoint"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_DURATION:
        return f"${name}->duration"
    if t in [xdr.SCSpecType.SC_SPEC_TYPE_U128, xdr.SCSpecType.SC_SPEC_TYPE_I128]:
        return f"gmp_strval(${name}->toBigInt())"
    if t in [xdr.SCSpecType.SC_SPEC_TYPE_U256, xdr.SCSpecType.SC_SPEC_TYPE_I256]:
        return f"gmp_strval(${name}->toBigInt())"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES:
        return f"${name}->bytes->getValue()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_STRING:
        return f"${name}->str"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_SYMBOL:
        return f"${name}->sym"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_ADDRESS:
        return f"Address::fromXdrSCVal(${name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MUXED_ADDRESS:
        return f"Address::fromXdrSCVal(${name})"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_OPTION:
        inner = from_scval(td.option.value_type, name, class_name)
        return f"(${name}->type !== XdrSCValType::SCV_VOID ? {inner} : null)"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_RESULT:
        ok_t = td.result.ok_type
        return from_scval(ok_t, name, class_name)
    if t == xdr.SCSpecType.SC_SPEC_TYPE_VEC:
        element_conversion = from_scval(td.vec.element_type, "item", class_name)
        return f"array_map(fn($item) => {element_conversion}, ${name}->vec)"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_MAP:
        key_conv = from_scval(td.map.key_type, "entry->key", class_name)
        val_conv = from_scval(td.map.value_type, "entry->val", class_name)
        return f"array_combine(array_map(fn($entry) => {key_conv}, ${name}->map), array_map(fn($entry) => {val_conv}, ${name}->map))"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_TUPLE:
        if len(td.tuple.value_types) == 0:
            return "null"
        conversions = [from_scval(td.tuple.value_types[i], f"{name}->vec[{i}]", class_name) for i in range(len(td.tuple.value_types))]
        return f"[{', '.join(conversions)}]"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_BYTES_N:
        return f"${name}->bytes->getValue()"
    if t == xdr.SCSpecType.SC_SPEC_TYPE_UDT:
        udt_name = td.udt.name.decode()
        return f"{prefixed_type_name(udt_name, class_name)}::fromSCVal(${name})"
    raise NotImplementedError(f"Unsupported SCValType: {t}")


def render_info():
    """Generate file header comment."""
    return f"""<?php

/**
 * This file was generated by stellar_contract_bindings v{stellar_contract_bindings_version}
 * and stellar_sdk v{stellar_sdk_version}.
 * 
 * @generated
 */
"""


def render_imports(namespace: str = "GeneratedContracts"):
    """Generate PHP namespace and use statements."""
    template = """
declare(strict_types=1);

namespace {{ namespace }};

use Exception;
use GuzzleHttp\\Exception\\GuzzleException;
use Soneso\\StellarSDK\\Crypto\\KeyPair;
use Soneso\\StellarSDK\\Soroban\\Address;
use Soneso\\StellarSDK\\Soroban\\Contract\\AssembledTransaction;
use Soneso\\StellarSDK\\Soroban\\Contract\\ClientOptions;
use Soneso\\StellarSDK\\Soroban\\Contract\\ContractSpec;
use Soneso\\StellarSDK\\Soroban\\Contract\\MethodOptions;
use Soneso\\StellarSDK\\Soroban\\Contract\\SorobanClient;
use Soneso\\StellarSDK\\Xdr\\XdrSCMapEntry;
use Soneso\\StellarSDK\\Xdr\\XdrSCVal;
use Soneso\\StellarSDK\\Xdr\\XdrSCValType;
"""
    return Template(template).render(namespace=namespace)


def render_enum(entry: xdr.SCSpecUDTEnumV0, class_name: str):
    """Generate PHP enum class."""
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/**
{%- if entry.doc %}
 * {{ entry.doc.decode() }}
{%- else %}
 * Generated enum {{ type_name }}
{%- endif %}
 */
enum {{ type_name }}: int
{
    {%- for case in entry.cases %}
    case {{ case.name.decode() }} = {{ case.value.uint32 }};
    {%- endfor %}

    public function toSCVal(): XdrSCVal
    {
        return XdrSCVal::forU32($this->value);
    }

    public static function fromSCVal(XdrSCVal $val): self
    {
        return self::from($val->u32);
    }
}
"""
    return Template(template).render(entry=entry, type_name=type_name)


def render_error_enum(entry: xdr.SCSpecUDTErrorEnumV0, class_name: str):
    """Generate PHP error enum class."""
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/**
{%- if entry.doc %}
 * {{ entry.doc.decode() }} (Error enum)
{%- else %}
 * Generated error enum {{ type_name }}
{%- endif %}
 */
enum {{ type_name }}Error: int
{
    {%- for case in entry.cases %}
    case {{ case.name.decode() }} = {{ case.value.uint32 }};
    {%- endfor %}

    public function toSCVal(): XdrSCVal
    {
        return XdrSCVal::forU32($this->value);
    }

    public static function fromSCVal(XdrSCVal $val): self
    {
        return self::from($val->u32);
    }
}
"""
    return Template(template).render(entry=entry, type_name=type_name)


def render_struct(entry: xdr.SCSpecUDTStructV0, class_name: str):
    """Generate PHP class for struct."""
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/**
{%- if entry.doc %}
 * {{ entry.doc.decode() }}
{%- else %}
 * Generated struct {{ type_name }}
{%- endif %}
 */
class {{ type_name }}
{
    {%- for field in entry.fields %}
    public {{ to_php_type(field.type, False, class_name) }} ${{ escape_keyword(field.name.decode()) }};
    {%- endfor %}

    public function __construct(
        {%- for field in entry.fields %}
        {{ to_php_type(field.type, False, class_name) }} ${{ escape_keyword(field.name.decode()) }}{% if not loop.last %},{% endif %}
        {%- endfor %}
    ) {
        {%- for field in entry.fields %}
        $this->{{ escape_keyword(field.name.decode()) }} = ${{ escape_keyword(field.name.decode()) }};
        {%- endfor %}
    }

    public function toSCVal(): XdrSCVal
    {
        $mapEntries = [];
        {%- for field in entry.fields %}
        $mapEntries[] = new XdrSCMapEntry(
            XdrSCVal::forSymbol('{{ field.name.decode() }}'),
            {{ to_scval(field.type, '$this->' ~ escape_keyword(field.name.decode()), class_name) }}
        );
        {%- endfor %}
        return XdrSCVal::forMap($mapEntries);
    }

    public static function fromSCVal(XdrSCVal $val): self
    {
        $map = [];
        foreach ($val->map as $entry) {
            $map[$entry->key->sym] = $entry->val;
        }
        return new self(
            {%- for field in entry.fields %}
            {{ escape_keyword(field.name.decode()) }}: {{ from_scval(field.type, 'map["' ~ field.name.decode() ~ '"]', class_name) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        );
    }
}
"""
    return Template(template).render(
        entry=entry,
        type_name=type_name,
        class_name=class_name,
        to_php_type=to_php_type,
        to_scval=to_scval,
        from_scval=from_scval,
        escape_keyword=escape_keyword
    )


def render_tuple_struct(entry: xdr.SCSpecUDTStructV0, class_name: str):
    """Generate PHP class for tuple struct."""
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/**
{%- if entry.doc %}
 * {{ entry.doc.decode() }}
{%- else %}
 * Generated tuple struct {{ type_name }}
{%- endif %}
 */
class {{ type_name }}
{
    public array $value;

    /**
     * @param array{%- for f in entry.fields %}{% if loop.first %}<{% endif %}{{ to_php_type(f.type, False, class_name) }}{% if not loop.last %}, {% else %}>{% endif %}{% endfor %} $value
     */
    public function __construct(array $value)
    {
        $this->value = $value;
    }

    public function toSCVal(): XdrSCVal
    {
        return XdrSCVal::forVec([
            {%- for f in entry.fields %}
            {{ to_scval(f.type, '$this->value[' ~ f.name.decode() ~ ']', class_name) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        ]);
    }

    public static function fromSCVal(XdrSCVal $val): self
    {
        $elements = $val->vec;
        return new self([
            {%- for f in entry.fields %}
            {{ from_scval(f.type, 'elements[' ~ f.name.decode() ~ ']', class_name) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        ]);
    }
}
"""
    return Template(template).render(
        entry=entry,
        type_name=type_name,
        class_name=class_name,
        to_php_type=to_php_type,
        to_scval=to_scval,
        from_scval=from_scval
    )


def render_union(entry: xdr.SCSpecUDTUnionV0, class_name: str):
    """Generate PHP class for union."""
    type_name = prefixed_type_name(entry.name.decode(), class_name)
    template = """
/**
{%- if entry.doc %}
 * {{ entry.doc.decode() }}
{%- else %}
 * Generated union {{ type_name }}
{%- endif %}
 */
class {{ type_name }}
{
    public const {% for case in entry.cases -%}
    {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 -%}
    {{ case.void_case.name.decode().upper() }} = '{{ case.void_case.name.decode() }}'
    {%- else -%}
    {{ case.tuple_case.name.decode().upper() }} = '{{ case.tuple_case.name.decode() }}'
    {%- endif -%}
    {%- if not loop.last %}, {% else %};{% endif -%}
    {%- endfor %}

    public string $kind;
    {%- for case in entry.cases %}
    {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
    {%- if len(case.tuple_case.type) == 1 %}
    public {{ to_php_type(case.tuple_case.type[0], nullable=True, class_name=class_name) }} ${{ camel_to_snake(case.tuple_case.name.decode()) }} = null;
    {%- else %}
    public ?array ${{ camel_to_snake(case.tuple_case.name.decode()) }} = null;
    {%- endif %}
    {%- endif %}
    {%- endfor %}

    public function __construct(string $kind{% for case in entry.cases -%}
    {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 -%}
    {%- if len(case.tuple_case.type) == 1 -%}
    , ?{{ to_php_type(case.tuple_case.type[0], False, class_name) }} ${{ camel_to_snake(case.tuple_case.name.decode()) }} = null
    {%- else -%}
    , ?array ${{ camel_to_snake(case.tuple_case.name.decode()) }} = null
    {%- endif -%}
    {%- endif -%}
    {%- endfor %})
    {
        $this->kind = $kind;
        {%- for case in entry.cases %}
        {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_TUPLE_V0 %}
        $this->{{ camel_to_snake(case.tuple_case.name.decode()) }} = ${{ camel_to_snake(case.tuple_case.name.decode()) }};
        {%- endif %}
        {%- endfor %}
    }

    public function toSCVal(): XdrSCVal
    {
        switch ($this->kind) {
            {%- for case in entry.cases %}
            {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
            case self::{{ case.void_case.name.decode().upper() }}:
                return XdrSCVal::forVec([XdrSCVal::forSymbol($this->kind)]);
            {%- else %}
            case self::{{ case.tuple_case.name.decode().upper() }}:
                {%- if len(case.tuple_case.type) == 1 %}
                return XdrSCVal::forVec([
                    XdrSCVal::forSymbol($this->kind),
                    {{ to_scval(case.tuple_case.type[0], '$this->' ~ camel_to_snake(case.tuple_case.name.decode()), class_name) }}
                ]);
                {%- else %}
                return XdrSCVal::forVec([
                    XdrSCVal::forSymbol($this->kind),
                    {%- for i, t in enumerate(case.tuple_case.type) %}
                    {{ to_scval(t, '$this->' ~ camel_to_snake(case.tuple_case.name.decode()) ~ '[' ~ i|string ~ ']', class_name) }}{% if not loop.last %},{% endif %}
                    {%- endfor %}
                ]);
                {%- endif %}
            {%- endif %}
            {%- endfor %}
            default:
                throw new Exception("Invalid union kind: {$this->kind}");
        }
    }

    public static function fromSCVal(XdrSCVal $val): self
    {
        if ($val->vec === null || count($val->vec) < 1) {
            throw new Exception("Invalid union value: expected vec with at least 1 element");
        }
        
        $kind = $val->vec[0]->sym;
        
        switch ($kind) {
            {%- for case in entry.cases %}
            {%- if case.kind == xdr.SCSpecUDTUnionCaseV0Kind.SC_SPEC_UDT_UNION_CASE_VOID_V0 %}
            case '{{ case.void_case.name.decode() }}':
                return new self(self::{{ case.void_case.name.decode().upper() }});
            {%- else %}
            case '{{ case.tuple_case.name.decode() }}':
                {%- if len(case.tuple_case.type) == 1 %}
                if (count($val->vec) !== 2) {
                    throw new Exception("Invalid union value for {{ case.tuple_case.name.decode() }}: expected 2 elements");
                }
                return new self(
                    self::{{ case.tuple_case.name.decode().upper() }},
                    {{ camel_to_snake(case.tuple_case.name.decode()) }}: {{ from_scval(case.tuple_case.type[0], 'val->vec[1]', class_name) }}
                );
                {%- else %}
                if (count($val->vec) !== {{ len(case.tuple_case.type) + 1 }}) {
                    throw new Exception("Invalid union value for {{ case.tuple_case.name.decode() }}: expected {{ len(case.tuple_case.type) + 1 }} elements");
                }
                return new self(
                    self::{{ case.tuple_case.name.decode().upper() }},
                    {{ camel_to_snake(case.tuple_case.name.decode()) }}: [
                        {%- for i, t in enumerate(case.tuple_case.type) %}
                        {{ from_scval(t, 'val->vec[' ~ (i + 1)|string ~ ']', class_name) }}{% if not loop.last %},{% endif %}
                        {%- endfor %}
                    ]
                );
                {%- endif %}
            {%- endif %}
            {%- endfor %}
            default:
                throw new Exception("Unknown union kind: $kind");
        }
    }
}
"""
    return Template(template).render(
        entry=entry,
        type_name=type_name,
        class_name=class_name,
        to_php_type=to_php_type,
        to_scval=to_scval,
        from_scval=from_scval,
        xdr=xdr,
        len=len,
        camel_to_snake=camel_to_snake,
        enumerate=enumerate
    )


def render_client(entries: List[xdr.SCSpecFunctionV0], contract_name: str):
    """Generate PHP client class."""
    template = '''
/**
 * Generated contract client for {{ contract_name }}
 */
class {{ contract_name }}
{
    /**
     * The underlying SorobanClient instance
     * @var SorobanClient
     */
    private SorobanClient $client;

    /**
     * Private constructor that wraps a SorobanClient
     * @param SorobanClient $client
     */
    private function __construct(SorobanClient $client)
    {
        $this->client = $client;
    }

    /**
     * Creates a new {{ contract_name }} for the given contract ID
     * @param ClientOptions $options Client options for the contract
     * @return {{ contract_name }}
     * @throws Exception
     * @throws GuzzleException
     */
    public static function forClientOptions(ClientOptions $options): self
    {
        $client = SorobanClient::forClientOptions($options);
        return new self($client);
    }

    /**
     * Gets the contract ID
     * @return string
     */
    public function getContractId(): string
    {
        return $this->client->getContractId();
    }

    /**
     * Gets the client options
     * @return ClientOptions
     */
    public function getOptions(): ClientOptions
    {
        return $this->client->getOptions();
    }

    /**
     * Gets the contract specification
     * @return ContractSpec
     */
    public function getContractSpec(): ContractSpec
    {
        return $this->client->getContractSpec();
    }
    {%- for entry in entries %}

    /**
    {%- if entry.doc %}
     * {{ entry.doc.decode() }}
    {%- else %}
     * Invoke the {{ entry.name.sc_symbol.decode() }} method
    {%- endif %}
     *
    {%- for param in entry.inputs %}
     * @param {{ to_php_type(param.type, False, contract_name) }} ${{ escape_keyword(param.name.decode()) }}
    {%- endfor %}
     * @param MethodOptions|null $methodOptions Options for transaction
     * @return {{ parse_result_type(entry.outputs) }}
     * @throws Exception
     * @throws GuzzleException
     */
    public function {{ snake_to_camel(entry.name.sc_symbol.decode()) }}(
        {%- for param in entry.inputs %}
        {{ to_php_type(param.type, False, contract_name) }} ${{ escape_keyword(param.name.decode()) }},
        {%- endfor %}
        ?MethodOptions $methodOptions = null
    ){{ return_type_hint(entry.outputs) }} {
        $args = [
            {%- for param in entry.inputs %}
            {{ to_scval(param.type, escape_keyword(param.name.decode()), contract_name) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        ];
        
        $result = $this->client->invokeMethod(
            name: '{{ entry.name.sc_symbol.decode() }}',
            args: $args,
            methodOptions: $methodOptions
        );
        {%- if len(entry.outputs) > 0 %}
        return {{ parse_result_conversion(entry.outputs) }};
        {%- endif %}
    }

    /**
     * Build an AssembledTransaction for the {{ entry.name.sc_symbol.decode() }} method.
     * This is useful if you need to manipulate the transaction before signing and sending.
     *
    {%- for param in entry.inputs %}
     * @param {{ to_php_type(param.type, False, contract_name) }} ${{ escape_keyword(param.name.decode()) }}
    {%- endfor %}
     * @param MethodOptions|null $methodOptions Options for transaction
     * @return AssembledTransaction
     * @throws Exception
     * @throws GuzzleException
     */
    public function build{{ snake_to_pascal(entry.name.sc_symbol.decode()) }}Tx(
        {%- for param in entry.inputs %}
        {{ to_php_type(param.type, False, contract_name) }} ${{ escape_keyword(param.name.decode()) }},
        {%- endfor %}
        ?MethodOptions $methodOptions = null
    ): AssembledTransaction {
        $args = [
            {%- for param in entry.inputs %}
            {{ to_scval(param.type, escape_keyword(param.name.decode()), contract_name) }}{% if not loop.last %},{% endif %}
            {%- endfor %}
        ];
        
        return $this->client->buildInvokeMethodTx(
            name: '{{ entry.name.sc_symbol.decode() }}',
            args: $args,
            methodOptions: $methodOptions
        );
    }
    {%- endfor %}
}
'''
    
    def parse_result_type(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return "void"
        elif len(output) == 1:
            return to_php_type(output[0], False, contract_name)
        else:
            return "array"
    
    def return_type_hint(output: List[xdr.SCSpecTypeDef]):
        result_type = parse_result_type(output)
        if result_type:
            # Add colon and return type
            return f": {result_type}"
        else:
            return ""
    
    def parse_result_conversion(output: List[xdr.SCSpecTypeDef]):
        if len(output) == 0:
            return ""
        elif len(output) == 1:
            return from_scval(output[0], "result", contract_name)
        else:
            conversions = [from_scval(output[i], f"result[{i}]", contract_name) for i in range(len(output))]
            return f"[{', '.join(conversions)}]"
    
    return Template(template).render(
        entries=entries,
        contract_name=contract_name,
        to_php_type=to_php_type,
        to_scval=to_scval,
        from_scval=from_scval,
        parse_result_type=parse_result_type,
        return_type_hint=return_type_hint,
        parse_result_conversion=parse_result_conversion,
        escape_keyword=escape_keyword,
        snake_to_camel=snake_to_camel,
        snake_to_pascal=snake_to_pascal,
        len=len
    )


def generate_binding(specs: List[xdr.SCSpecEntry], namespace: str = "GeneratedContracts", contract_name: str = "Contract") -> str:
    """Generate complete PHP binding file."""
    generated = []
    generated.append(render_info())
    generated.append(render_imports(namespace))
    
    # Generate types
    for spec in specs:
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ENUM_V0:
            generated.append(render_enum(spec.udt_enum_v0, contract_name))
        elif spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_ERROR_ENUM_V0:
            generated.append(render_error_enum(spec.udt_error_enum_v0, contract_name))
        elif spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_STRUCT_V0:
            if is_tuple_struct(spec.udt_struct_v0):
                generated.append(render_tuple_struct(spec.udt_struct_v0, contract_name))
            else:
                generated.append(render_struct(spec.udt_struct_v0, contract_name))
        elif spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_UDT_UNION_V0:
            generated.append(render_union(spec.udt_union_v0, contract_name))
    
    # Generate client
    function_specs: List[xdr.SCSpecFunctionV0] = [
        spec.function_v0
        for spec in specs
        if spec.kind == xdr.SCSpecEntryKind.SC_SPEC_ENTRY_FUNCTION_V0
        and not spec.function_v0.name.sc_symbol.decode().startswith("__")
    ]
    
    if function_specs:
        generated.append(render_client(function_specs, contract_name))
    
    return "\n".join(generated)


@click.command(name="php")
@click.option(
    "--contract-id", required=True, help="The contract ID to generate bindings for"
)
@click.option(
    "--rpc-url", default="https://mainnet.sorobanrpc.com", help="Soroban RPC URL"
)
@click.option(
    "--output",
    default=None,
    help="Output directory for generated bindings, defaults to current directory",
)
@click.option(
    "--namespace",
    default="GeneratedContracts",
    help="PHP namespace for generated classes",
)
@click.option(
    "--class-name",
    default="ContractClient",
    help="Name for the generated client class",
)
def command(contract_id: str, rpc_url: str, output: str, namespace: str, class_name: str):
    """Generate PHP bindings for a Soroban contract"""
    if not StrKey.is_valid_contract(contract_id):
        click.echo(f"Invalid contract ID: {contract_id}", err=True)
        raise click.Abort()

    # Use current directory if output is not specified
    if output is None:
        output = os.getcwd()
    
    try:
        specs = get_specs_by_contract_id(contract_id, rpc_url)
    except Exception as e:
        click.echo(f"Get contract specs failed: {e}", err=True)
        raise click.Abort()

    click.echo("Generating PHP bindings")
    generated = generate_binding(specs, namespace=namespace, contract_name=class_name)
    
    if not os.path.exists(output):
        os.makedirs(output)
    
    output_path = os.path.join(output, f"{class_name}.php")
    with open(output_path, "w") as f:
        f.write(generated)
    
    click.echo(f"Generated PHP bindings to {output_path}")


if __name__ == "__main__":
    command()